# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import signal
from threading import Lock
from time import sleep

from ovos_plugin_manager.templates.audio import AudioBackend
from ovos_bus_client.message import Message
from ovos_utils.log import LOG
from ovos_utils.sound import play_audio
import mimetypes
import re
from requests import Session


def find_mime(path):
    mime = None
    if path.startswith('http'):
        response = Session().head(path, allow_redirects=True)
        if 200 <= response.status_code < 300:
            mime = response.headers['content-type']
    if not mime:
        mime = mimetypes.guess_type(path)[0]
    # Remove any http address arguments
    if not mime:
        mime = mimetypes.guess_type(re.sub(r'\?.*$', '', path))[0]

    if mime:
        return mime.split('/')
    else:
        return (None, None)


class SimpleAudioService(AudioBackend):
    """
        Simple Audio backend for both mpg123 and the ogg123 player.
        This one is rather limited and only implements basic usage.
    """

    def __init__(self, config, bus, name='simple'):
        super().__init__(config, bus)
        self.config = config
        self.process = None
        self.bus = bus
        self.name = name
        self._stop_signal = False
        self._is_playing = False
        self._paused = False
        self.tracks = []
        self.index = 0
        self.supports_mime_hints = True
        mimetypes.init()
        self.track_lock = Lock()

        self.bus.on('SimpleAudioServicePlay', self._play)

    def supported_uris(self):
        return ['file', 'http']

    def clear_list(self):
        with self.track_lock:
            self.tracks = []

    def add_list(self, tracks):
        with self.track_lock:
            self.tracks += tracks
            LOG.info("Track list is " + str(tracks))

    def _get_track(self, track_data):
        if isinstance(track_data, list):
            track = track_data[0]
            mime = track_data[1]
            mime = mime.split('/')
        else:  # Assume string
            track = track_data
            mime = find_mime(track)
        return track, mime

    def _play(self, message):
        """Implementation specific async method to handle playback.

        This allows mpg123 service to use the next method as well
        as basic play/stop.
        """
        LOG.info('SimpleAudioService._play')

        # Stop any existing audio playback
        self._stop_running_process()

        repeat = message.data.get('repeat', False)
        self._is_playing = True
        self._paused = False
        with self.track_lock:
            if len(self.tracks) > self.index:
                track, mime = self._get_track(self.tracks[self.index])
            else:
                return

        LOG.debug('Mime info: {}'.format(mime))

        # Indicate to audio service which track is being played
        if self._track_start_callback:
            self._track_start_callback(track)

        # Replace file:// uri's with normal paths
        track = track.replace('file://', '')
        try:
            self.process = play_audio(track)
        except FileNotFoundError as e:
            LOG.error('Couldn\'t play audio, {}'.format(repr(e)))
            self.process = None
        except Exception as e:
            LOG.exception(repr(e))
            self.process = None

        # Wait for completion or stop request
        while (self._is_process_running() and not self._stop_signal):
            sleep(0.25)

        if self._stop_signal:
            self._stop_running_process()
            self._is_playing = False
            self._paused = False
            return
        else:
            self.process = None

        # if there are more tracks available play next
        self.index += 1
        with self.track_lock:
            if self.index < len(self.tracks) or repeat:
                if self.index >= len(self.tracks):
                    self.index = 0
                self.bus.emit(Message('SimpleAudioServicePlay',
                                      {'repeat': repeat}))
            else:
                self._track_start_callback(None)
                self._is_playing = False
                self._paused = False

    def play(self, repeat=False):
        LOG.info('Call SimpleAudioServicePlay')
        self.index = 0
        self.bus.emit(Message('SimpleAudioServicePlay', {'repeat': repeat}))

    def stop(self):
        LOG.info('SimpleAudioServiceStop')
        self._stop_signal = True
        while self._is_playing:
            sleep(0.1)
        self._stop_signal = False

    def _pause(self):
        """ Pauses playback if possible.

            Returns: (bool) New paused status:
        """
        if self.process:
            # Suspend the playback process
            self.process.send_signal(signal.SIGSTOP)
            return True  # After pause the service is paused
        else:
            return False

    def pause(self):
        if not self._paused:
            self._paused = self._pause()

    def _resume(self):
        """ Resumes playback if possible.

            Returns: (bool) New paused status:
        """
        if self.process:
            # Resume the playback process
            self.process.send_signal(signal.SIGCONT)
            return False  # After resume the service is no longer paused
        else:
            return True

    def resume(self):
        if self._paused:
            # Resume the playback process
            self._paused = self._resume()

    def next(self):
        # Terminate process to continue to next
        self._stop_running_process()

    def previous(self):
        pass

    def lower_volume(self):
        if not self._paused:
            self._pause()  # poor-man's ducking

    def restore_volume(self):
        if not self._paused:
            self._resume()  # poor-man's unducking

    def _is_process_running(self):
        return self.process and self.process.poll() is None

    def _stop_running_process(self):
        if self._is_process_running():
            if self._paused:
                # The child process must be "unpaused" in order to be stopped
                self._resume()
            self.process.terminate()
            countdown = 10
            while self._is_process_running() and countdown > 0:
                sleep(0.1)
                countdown -= 1

            if self._is_process_running():
                # Failed to shutdown when asked nicely.  Force the issue.
                LOG.debug("Killing currently playing audio...")
                self.process.kill()
        self.process = None

    # mandatory abstract methods
    def get_track_length(self) -> int:
        return 0

    def get_track_position(self) -> int:
        return 0

    def set_track_position(self, milliseconds):
        pass


def load_service(base_config, bus):
    backends = base_config.get('backends', [])
    services = [(b, backends[b]) for b in backends
                if backends[b]['type'] == 'simple' and
                backends[b].get('active', False)]
    instances = [SimpleAudioService(s[1], bus, s[0]) for s in services]
    return instances



# backwards compat imports
from ovos_audio.tts import TTSFactory
from ovos_plugin_manager.tts import load_tts_plugin
from ovos_plugin_manager.templates.tts import TTS, RemoteTTS, \
    RemoteTTSException, RemoteTTSTimeoutException


import os
import re
from os.path import dirname
from threading import Event
from typing import Optional, List

from langcodes import closest_match

from ovos_bus_client.message import Message
from ovos_bus_client.session import SessionManager
from ovos_config.config import Configuration
from ovos_plugin_manager.templates.pipeline import PipelineMatch, PipelinePlugin
from ovos_utils import flatten_list
from ovos_utils.bracket_expansion import expand_options
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.log import LOG
from ovos_utils.parse import match_one


class StopService(PipelinePlugin):
    """Intent Service thats handles stopping skills."""

    def __init__(self, bus):
        self.bus = bus
        self._voc_cache = {}
        self.load_resource_files()
        super().__init__(config=Configuration().get("skills", {}).get("stop") or {})

    def load_resource_files(self):
        base = f"{dirname(__file__)}/locale"
        for lang in os.listdir(base):
            lang2 = standardize_lang_tag(lang)
            self._voc_cache[lang2] = {}
            for f in os.listdir(f"{base}/{lang}"):
                with open(f"{base}/{lang}/{f}", encoding="utf-8") as fi:
                    lines = [expand_options(l) for l in fi.read().split("\n")
                             if l.strip() and not l.startswith("#")]
                    n = f.split(".", 1)[0]
                    self._voc_cache[lang2][n] = flatten_list(lines)

    @staticmethod
    def get_active_skills(message: Optional[Message] = None) -> List[str]:
        """Active skill ids ordered by converse priority
        this represents the order in which stop will be called

        Returns:
            active_skills (list): ordered list of skill_ids
        """
        session = SessionManager.get(message)
        return [skill[0] for skill in session.active_skills]

    def _collect_stop_skills(self, message: Message) -> List[str]:
        """use the messagebus api to determine which skills can stop
        This includes all skills and external applications"""

        want_stop = []
        skill_ids = []

        active_skills = self.get_active_skills(message)

        if not active_skills:
            return want_stop

        event = Event()

        def handle_ack(msg):
            nonlocal event
            skill_id = msg.data["skill_id"]

            # validate the stop pong
            if all((skill_id not in want_stop,
                    msg.data.get("can_handle", True),
                    skill_id in active_skills)):
                want_stop.append(skill_id)

            if skill_id not in skill_ids:  # track which answer we got
                skill_ids.append(skill_id)

            if all(s in skill_ids for s in active_skills):
                # all skills answered the ping!
                event.set()

        self.bus.on("skill.stop.pong", handle_ack)

        # ask skills if they can stop
        for skill_id in active_skills:
            self.bus.emit(message.forward(f"{skill_id}.stop.ping",
                                          {"skill_id": skill_id}))

        # wait for all skills to acknowledge they can stop
        event.wait(timeout=0.5)

        self.bus.remove("skill.stop.pong", handle_ack)
        return want_stop or active_skills

    def stop_skill(self, skill_id: str, message: Message) -> bool:
        """Tell a skill to stop anything it's doing,
        taking into account the message Session

        Args:
            skill_id: skill to query.
            message (Message): message containing interaction info.

        Returns:
            handled (bool): True if handled otherwise False.
        """
        stop_msg = message.reply(f"{skill_id}.stop")
        result = self.bus.wait_for_response(stop_msg, f"{skill_id}.stop.response")
        if result and 'error' in result.data:
            error_msg = result.data['error']
            LOG.error(f"{skill_id}: {error_msg}")
            return False
        elif result is not None:
            return result.data.get('result', False)

    def match_stop_high(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """If utterance is an exact match for "stop" , run before intent stage

        Args:
            utterances (list):  list of utterances
            lang (string):      4 letter ISO language code
            message (Message):  message to use to generate reply

        Returns:
            PipelineMatch if handled otherwise None.
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return None

        sess = SessionManager.get(message)

        # we call flatten in case someone is sending the old style list of tuples
        utterance = flatten_list(utterances)[0]

        is_stop = self.voc_match(utterance, 'stop', exact=True, lang=lang)
        is_global_stop = self.voc_match(utterance, 'global_stop', exact=True, lang=lang) or \
                         (is_stop and not len(self.get_active_skills(message)))

        conf = 1.0

        if is_global_stop:
            # emit a global stop, full stop anything OVOS is doing
            self.bus.emit(message.reply("mycroft.stop", {}))
            return PipelineMatch(handled=True,
                                 match_data={"conf": conf},
                                 skill_id=None,
                                 utterance=utterance)

        if is_stop:
            # check if any skill can stop
            for skill_id in self._collect_stop_skills(message):
                if skill_id in sess.blacklisted_skills:
                    LOG.debug(f"ignoring match, skill_id '{skill_id}' blacklisted by Session '{sess.session_id}'")
                    continue

                if self.stop_skill(skill_id, message):
                    return PipelineMatch(handled=True,
                                         match_data={"conf": conf},
                                         skill_id=skill_id,
                                         utterance=utterance)
        return None

    def match_stop_medium(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """ if "stop" intent is in the utterance,
        but it contains additional words not in .intent files

        Args:
            utterances (list):  list of utterances
            lang (string):      4 letter ISO language code
            message (Message):  message to use to generate reply

        Returns:
            PipelineMatch if handled otherwise None.
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return None

        # we call flatten in case someone is sending the old style list of tuples
        utterance = flatten_list(utterances)[0]

        is_stop = self.voc_match(utterance, 'stop', exact=False, lang=lang)
        if not is_stop:
            is_global_stop = self.voc_match(utterance, 'global_stop', exact=False, lang=lang) or \
                             (is_stop and not len(self.get_active_skills(message)))
            if not is_global_stop:
                return None

        return self.match_stop_low(utterances, lang, message)

    def _get_closest_lang(self, lang: str) -> Optional[str]:
        if self._voc_cache:
            lang = standardize_lang_tag(lang)
            closest, score = closest_match(lang, list(self._voc_cache.keys()))
            # https://langcodes-hickford.readthedocs.io/en/sphinx/index.html#distance-values
            # 0 -> These codes represent the same language, possibly after filling in values and normalizing.
            # 1- 3 -> These codes indicate a minor regional difference.
            # 4 - 10 -> These codes indicate a significant but unproblematic regional difference.
            if score < 10:
                return closest
        return None

    def match_stop_low(self, utterances: List[str], lang: str, message: Message) -> Optional[PipelineMatch]:
        """ before fallback_low , fuzzy match stop intent

        Args:
            utterances (list):  list of utterances
            lang (string):      4 letter ISO language code
            message (Message):  message to use to generate reply

        Returns:
            PipelineMatch if handled otherwise None.
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return None
        sess = SessionManager.get(message)
        # we call flatten in case someone is sending the old style list of tuples
        utterance = flatten_list(utterances)[0]

        conf = match_one(utterance, self._voc_cache[lang]['stop'])[1]
        if len(self.get_active_skills(message)) > 0:
            conf += 0.1
        conf = round(min(conf, 1.0), 3)

        if conf < self.config.get("min_conf", 0.5):
            return None

        # check if any skill can stop
        for skill_id in self._collect_stop_skills(message):
            if skill_id in sess.blacklisted_skills:
                LOG.debug(f"ignoring match, skill_id '{skill_id}' blacklisted by Session '{sess.session_id}'")
                continue

            if self.stop_skill(skill_id, message):
                return PipelineMatch(handled=True,
                                     # emit instead of intent message
                                     match_data={"conf": conf},
                                     skill_id=skill_id, utterance=utterance)

        # emit a global stop, full stop anything OVOS is doing
        self.bus.emit(message.reply("mycroft.stop", {}))
        return PipelineMatch(handled=True,
                             # emit instead of intent message {"conf": conf},
                             match_data={"conf": conf},
                             skill_id=None,
                             utterance=utterance)

    def voc_match(self, utt: str, voc_filename: str, lang: str,
                  exact: bool = False):
        """
        Determine if the given utterance contains the vocabulary provided.

        By default the method checks if the utterance contains the given vocab
        thereby allowing the user to say things like "yes, please" and still
        match against "Yes.voc" containing only "yes". An exact match can be
        requested.

        The method first checks in the current Skill's .voc files and secondly
        in the "res/text" folder of mycroft-core. The result is cached to
        avoid hitting the disk each time the method is called.

        Args:
            utt (str): Utterance to be tested
            voc_filename (str): Name of vocabulary file (e.g. 'yes' for
                                'res/text/en-us/yes.voc')
            lang (str): Language code, defaults to self.lang
            exact (bool): Whether the vocab must exactly match the utterance

        Returns:
            bool: True if the utterance has the given vocabulary it
        """
        lang = self._get_closest_lang(lang)
        if lang is None:  # no vocs registered for this lang
            return False

        _vocs = self._voc_cache[lang].get(voc_filename) or []

        if utt and _vocs:
            if exact:
                # Check for exact match
                return any(i.strip() == utt
                           for i in _vocs)
            else:
                # Check for matches against complete words
                return any([re.match(r'.*\b' + i + r'\b.*', utt)
                            for i in _vocs])
        return False



from ovos_backend_client.api import DeviceApi
from ovos_backend_client.pairing import is_paired
from ovos_utils.log import LOG
from ovos_backend_client.settings import SeleneSkillsManifest
from ovos_config import Configuration
# backwards compat import - do not delete
from mycroft.deprecated.skills.skill_updater import SkillUpdater


class SeleneSkillManifestUploader:
    """Class facilitating skill manifest upload."""

    def __init__(self):
        super().__init__()
        self.api = DeviceApi()
        self.config = Configuration()
        self.skill_manifest = SeleneSkillsManifest(self.api)
        self.post_manifest(True)

    @property
    def installed_skills_file_path(self):
        """Property representing the path of the installed skills file."""
        return self.skill_manifest.path

    def post_manifest(self, reload_skills_manifest=False):
        """Post the manifest of the device's skills to the backend."""
        upload_allowed = self.config['skills'].get('upload_skill_manifest')
        if upload_allowed and is_paired():
            if reload_skills_manifest:
                self.skill_manifest.clear()
                self.skill_manifest.scan_skills()
            try:
                self.api.upload_skills_data(self.skill_manifest)
            except Exception:
                LOG.error('Could not upload skill manifest')



# This submodule has been deprecated, please import from ovos_backend_client directly
from mycroft.deprecated.api import Api, UUID, GeolocationApi, STTApi, DeviceApi
from ovos_backend_client.pairing import has_been_paired, is_paired, check_remote_pairing, \
    is_backend_disabled, requires_backend
from ovos_backend_client.exceptions import BackendDown, InternetDown



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
NOTE: this is dead code! do not use!
This file is only present to ensure backwards compatibility
in case someone is importing from here
This is only meant for 3rd party code expecting ovos-core
to be a drop in replacement for mycroft-core
"""

from threading import Lock
from ovos_bus_client import MessageBusClient, Message
from ovos_utils.log import LOG

bus: MessageBusClient = None
tts = None
tts_hash = None
lock = Lock()
config = None
mimic_fallback_obj = None


def _get_messagebus():
    global bus
    if not bus:
        bus = MessageBusClient()
        bus.run_in_thread()
    return bus


def handle_speak(event):
    LOG.warning("speech.handle_speak has been deprecated!")
    LOG.error("speak message not handled")


def mute_and_speak(utterance, ident, listen=False):
    LOG.warning("speech.mute_and_speak has been deprecated!")
    bus.emit(Message("speak", {"utterance": utterance,
                               "expect_response": listen}, {"ident": ident}))


def mimic_fallback_tts(utterance, ident, listen):
    """
    DEPRECATED: use execute_fallback_tts instead
    This method is only kept around for backwards api compat
    """
    LOG.warning("speech.mimic_fallback_tts is deprecated! "
                "use audio.service.service.execute_fallback_tts instead")
    mute_and_speak(utterance, ident, listen)


def handle_stop(event):
    LOG.warning("speech.handle_stop has been deprecated!")
    _get_messagebus().emit(Message("mycroft.stop"))


def init(_):
    LOG.warning("speech.init has been deprecated!")


def shutdown():
    """Shutdown the audio service cleanly.
    Stop any playing audio and make sure threads are joined correctly.
    """
    LOG.warning("speech.shutdown has been deprecated!")



# backwards compat imports
from ovos_listener.hotword_factory import HotWordFactory
from ovos_plugin_manager.wakewords import load_wake_word_plugin, find_wake_word_plugins
from ovos_plugin_manager.templates.hotwords import HotWordEngine


# backwards compat imports
from ovos_listener.listener import RecognizerLoop
from ovos_listener.mic import ListenerState, ListeningMode
from ovos_listener.service import SpeechService, on_ready, on_stopping, on_error
from ovos_utils.log import LOG


class SpeechClient(SpeechService):
    def __init__(self, *args, **kwargs):
        LOG.warning("SpeechClient has been renamed to SpeechService, it will be removed in 0.1.0")
        super().__init__(self, *args, **kwargs)




"""
NOTE: this is dead code! do not use!
This file is only present to ensure backwards compatibility
in case someone is importing from here
This is only meant for 3rd party code expecting ovos-core
to be a drop in replacement for mycroft-core
"""

import collections
import csv
import re
from os import walk
from os.path import splitext, join
from ovos_backend_client.pairing import is_paired
from ovos_bus_client.apis.enclosure import EnclosureAPI
from mycroft.util.format import expand_options
from ovos_utils.log import LOG
from ovos_workshop.intents import munge_regex, to_alnum

RASPBERRY_PI_PLATFORMS = ('mycroft_mark_1', 'picroft', 'mycroft_mark_2pi')

ONE_MINUTE = 60


def skill_is_blacklisted(skill):
    """DEPRECATED: do not use, method only for api backwards compatibility
    Logs a warning and returns False
    """
    # this is a internal msm helper
    # it should have been private
    # cant remove to keep api compatibility
    # nothing in the wild should be using this
    LOG.warning("skill_is_blacklisted is an internal method and has been deprecated. Stop using it!")
    return False


class DevicePrimer:
    """DEPRECATED: this class has been fully deprecated, stop using it!
    Only here to provide public api compatibility but it does absolutely nothing!
    """

    def __init__(self, message_bus_client, config=None):
        self.bus = message_bus_client
        self.platform = "unknown"
        self.enclosure = EnclosureAPI(self.bus)
        self.backend_down = False

    @property
    def is_paired(self):
        return is_paired()

    def prepare_device(self):
        """Internet dependent updates of various aspects of the device."""
        LOG.warning("DevicePrimer has been deprecated!")


def read_vocab_file(path):
    """ Read voc file.

        This reads a .voc file, stripping out empty lines comments and expand
        parentheses. It returns each line as a list of all expanded
        alternatives.

        Args:
            path (str): path to vocab file.

        Returns:
            List of Lists of strings.
    """
    LOG.warning("read_vocab_file is deprecated! "
                "use SkillResources class instead")
    vocab = []
    with open(path, 'r', encoding='utf8') as voc_file:
        for line in voc_file.readlines():
            if line.startswith('#') or line.strip() == '':
                continue
            vocab.append(expand_options(line.lower()))
    return vocab


def load_regex_from_file(path, skill_id):
    """Load regex from file
    The regex is sent to the intent handler using the message bus

    Args:
        path:       path to vocabulary file (*.voc)
        skill_id:   skill_id to the regex is tied to
    """
    LOG.warning("read_regex_from_file is deprecated! "
                "use SkillResources class instead")
    regexes = []
    if path.endswith('.rx'):
        with open(path, 'r', encoding='utf8') as reg_file:
            for line in reg_file.readlines():
                if line.startswith("#"):
                    continue
                LOG.debug('regex pre-munge: ' + line.strip())
                regex = munge_regex(line.strip(), skill_id)
                LOG.debug('regex post-munge: ' + regex)
                # Raise error if regex can't be compiled
                try:
                    re.compile(regex)
                    regexes.append(regex)
                except Exception as e:
                    LOG.warning(f'Failed to compile regex {regex}: {e}')

    return regexes


def load_vocabulary(basedir, skill_id):
    """Load vocabulary from all files in the specified directory.

    Args:
        basedir (str): path of directory to load from (will recurse)
        skill_id: skill the data belongs to
    Returns:
        dict with intent_type as keys and list of list of lists as value.
    """
    LOG.warning("load_vocabulary is deprecated! "
                "use SkillResources class instead")
    vocabs = {}
    for path, _, files in walk(basedir):
        for f in files:
            if f.endswith(".voc"):
                vocab_type = to_alnum(skill_id) + splitext(f)[0]
                vocs = read_vocab_file(join(path, f))
                if vocs:
                    vocabs[vocab_type] = vocs
    return vocabs


def load_regex(basedir, skill_id):
    """Load regex from all files in the specified directory.

    Args:
        basedir (str): path of directory to load from
        bus (messagebus emitter): messagebus instance used to send the vocab to
                                  the intent service
        skill_id (str): skill identifier
    """
    LOG.warning("load_regex is deprecated! "
                "use SkillResources class instead")
    regexes = []
    for path, _, files in walk(basedir):
        for f in files:
            if f.endswith(".rx"):
                regexes += load_regex_from_file(join(path, f), skill_id)
    return regexes


def read_value_file(filename, delim):
    """Read value file.

    The value file is a simple csv structure with a key and value.

    Args:
        filename (str): file to read
        delim (str): csv delimiter

    Returns:
        OrderedDict with results.
    """
    LOG.warning("read_value_file is deprecated! "
                "use SkillResources class instead")
    result = collections.OrderedDict()

    if filename:
        with open(filename) as f:
            reader = csv.reader(f, delimiter=delim)
            for row in reader:
                # skip blank or comment lines
                if not row or row[0].startswith("#"):
                    continue
                if len(row) != 2:
                    continue

                result[row[0]] = row[1]
    return result


def read_translated_file(filename, data):
    """Read a file inserting data.

    Args:
        filename (str): file to read
        data (dict): dictionary with data to insert into file

    Returns:
        list of lines.
    """
    LOG.warning("read_translated_file is deprecated! "
                "use SkillResources class instead")
    if filename:
        with open(filename) as f:
            text = f.read().replace('{{', '{').replace('}}', '}')
            return text.format(**data or {}).rstrip('\n').split('\n')
    else:
        return None



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Message bus service for mycroft-core

The message bus facilitates inter-process communication between mycroft-core
processes. It implements a websocket server so can also be used by external
systems to integrate with the Mycroft system.
"""
from ovos_bus_client.conf import load_message_bus_config
from ovos_messagebus.event_handler import MessageBusEventHandler
from ovos_utils.process_utils import reset_sigint_handler
from ovos_utils import create_daemon, wait_for_exit_signal
from ovos_utils.log import LOG, init_service_logger
from tornado import web, ioloop


def on_ready():
    LOG.info('Message bus service started!')


def on_error(e='Unknown'):
    LOG.info('Message bus failed to start ({})'.format(repr(e)))


def on_stopping():
    LOG.info('Message bus is shutting down...')


def main(ready_hook=on_ready, error_hook=on_error, stopping_hook=on_stopping):
    init_service_logger("bus")
    LOG.info('Starting message bus service...')
    reset_sigint_handler()
    config = load_message_bus_config()
    routes = [(config.route, MessageBusEventHandler)]
    application = web.Application(routes)
    application.listen(config.port, config.host)
    create_daemon(ioloop.IOLoop.instance().start)
    ready_hook()
    wait_for_exit_signal()
    stopping_hook()


if __name__ == "__main__":
    main()



"""
This module contains back compat imports only
All code here is in life support and available only for backwards api compatibility
Bug fixes will be made to maintain compatibility but are very low priority!
"""


# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Periodically run by skill manager to update skills and post the manifest."""

from combo_lock import ComboLock

from ovos_utils.file_utils import get_temp_path
from ovos_utils.log import LOG
from ovos_config.config import Configuration

ONE_HOUR = 3600
FIVE_MINUTES = 300  # number of seconds in a minute


class SkillUpdater:
    """
    DEPRECATED - most of this class is now useless and logs warnings only
    Please use SeleneSkillManifestUploader to post skills manifest
    Skill updates are no longer handled by core

    Class facilitating skill update / install actions.

    Arguments
        bus (MessageBusClient): Optional bus emitter Used to communicate
                                with the mycroft core system and handle
                                commands.
    """

    def __init__(self, bus=None):
        self.installed_skills = set()
        self.msm_lock = ComboLock(get_temp_path('mycroft-msm.lck'))
        self.install_retries = 0
        self.config = Configuration()
        update_interval = self.config['skills'].get('update_interval', 1.0)
        self.update_interval = int(update_interval) * ONE_HOUR
        self.dot_msm_path = "/tmp/.msm"
        self.next_download = 0
        self.default_skill_install_error = False

        self.post_manifest(True)
        if bus:
            LOG.warning("bus argument has been deprecated")

    @property
    def installed_skills_file_path(self):
        """Property representing the path of the installed skills file."""
        from mycroft.skills.skill_updater import SeleneSkillManifestUploader
        return SeleneSkillManifestUploader().skill_manifest.path

    @property
    def msm(self):
        """DEPRECATED: do not use, method only for api backwards compatibility
        Logs a warning and returns None
        """
        # unused but need to keep api backwards compatible
        # log a warning and move on
        LOG.warning("msm has been deprecated\n"
                    "DO NOT use self.msm property")
        return None

    @property
    def default_skill_names(self) -> tuple:
        """Property representing the default skills expected to be installed"""
        LOG.warning("msm has been deprecated\n"
                    "skill install/update is no longer handled by ovos-core")
        return ()

    def update_skills(self, quick=False):
        """DEPRECATED: do not use, method only for api backwards compatibility
        Logs a warning and returns True
        """
        LOG.warning("msm has been deprecated\n"
                    "skill install/update is no longer handled by ovos-core")
        return True

    def handle_not_connected(self):
        """"DEPRECATED: do not use, method only for api backwards compatibility
        Logs a warning
        """
        LOG.warning("msm has been deprecated\n"
                    "no update will be scheduled")

    def post_manifest(self, reload_skills_manifest=False):
        """Post the manifest of the device's skills to the backend."""
        from mycroft.skills.skill_updater import SeleneSkillManifestUploader
        uploader = SeleneSkillManifestUploader()
        uploader.post_manifest(reload_skills_manifest)

    def install_or_update(self, skill):
        """DEPRECATED: do not use, method only for api backwards compatibility
        Logs a warning
        """
        LOG.warning("msm has been deprecated\n"
                    f"{skill} will not be changed")

    def defaults_installed(self):
        """DEPRECATED: do not use, method only for api backwards compatibility
        Logs a warning and returns True
        """
        LOG.warning("msm has been deprecated\n"
                    "skill install/update is no longer handled by ovos-core")
        return True



from ovos_core.version import OVOS_VERSION_STR, OVOS_VERSION_BUILD, OVOS_VERSION_TUPLE, \
    OVOS_VERSION_ALPHA, VersionManager, check_version

# The following lines are replaced during the release process.
# START_VERSION_BLOCK
CORE_VERSION_MAJOR = 21
CORE_VERSION_MINOR = 2
CORE_VERSION_BUILD = 1

# END_VERSION_BLOCK
CORE_VERSION_TUPLE = (CORE_VERSION_MAJOR,
                      CORE_VERSION_MINOR,
                      CORE_VERSION_BUILD)
CORE_VERSION_STR = '.'.join(map(str, CORE_VERSION_TUPLE))



"""
This module contains back compat imports only
Speech client moved into mycroft.listener module
"""
from ovos_listener.hotword_factory import HotWordFactory
from ovos_plugin_manager.templates.hotwords import HotWordEngine




# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from ovos_listener.service import SpeechService, on_error, on_stopping, on_ready
from ovos_config.locale import setup_locale
from ovos_utils.log import init_service_logger
from ovos_utils.process_utils import reset_sigint_handler, PIDLock
from ovos_utils import wait_for_exit_signal


def main(ready_hook=on_ready, error_hook=on_error, stopping_hook=on_stopping,
         watchdog=lambda: None):
    init_service_logger("voice")
    reset_sigint_handler()
    PIDLock("voice")
    setup_locale()
    service = SpeechService(on_ready=ready_hook,
                            on_error=error_hook,
                            on_stopping=stopping_hook,
                            watchdog=watchdog)
    service.daemon = True
    service.start()
    wait_for_exit_signal()


if __name__ == "__main__":
    main()



# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Contains simple tools for performing audio related tasks such as playback
of audio, recording and listing devices.
"""
import re
from ovos_utils.log import LOG
from ovos_utils.sound import play_audio
try:
    import pyaudio
except ImportError:
    pyaudio = None


def play_audio_file(uri: str, environment=None):
    """Play an audio file.

    This wraps the other play_* functions, choosing the correct one based on
    the file extension. The function will return directly and play the file
    in the background.

    Args:
        uri:    uri to play
        environment (dict): optional environment for the subprocess call

    Returns: subprocess.Popen object. None if the format is not supported or
             an error occurs playing the file.
    """
    return play_audio(uri, environment=environment)


def play_wav(uri, environment=None):
    """Play a wav-file.

    This will use the application specified in the mycroft config
    and play the uri passed as argument. The function will return directly
    and play the file in the background.

    Args:
        uri:    uri to play
        environment (dict): optional environment for the subprocess call

    Returns: subprocess.Popen object or None if operation failed
    """
    return play_audio(uri, environment=environment)


def play_mp3(uri, environment=None):
    """Play a mp3-file.

    This will use the application specified in the mycroft config
    and play the uri passed as argument. The function will return directly
    and play the file in the background.

    Args:
        uri:    uri to play
        environment (dict): optional environment for the subprocess call

    Returns: subprocess.Popen object or None if operation failed
    """
    return play_audio(uri, environment=environment)


def play_ogg(uri, environment=None):
    """Play an ogg-file.

    This will use the application specified in the mycroft config
    and play the uri passed as argument. The function will return directly
    and play the file in the background.

    Args:
        uri:    uri to play
        environment (dict): optional environment for the subprocess call

    Returns: subprocess.Popen object, or None if operation failed
    """
    return play_audio(uri, environment=environment)


def find_input_device(device_name):
    """Find audio input device by name.

    Args:
        device_name: device name or regex pattern to match

    Returns: device_index (int) or None if device wasn't found
    """
    if pyaudio is None:
        raise ImportError("pyaudio not installed")
    LOG.info('Searching for input device: {}'.format(device_name))
    LOG.debug('Devices: ')
    pa = pyaudio.PyAudio()
    pattern = re.compile(device_name)
    for device_index in range(pa.get_device_count()):
        dev = pa.get_device_info_by_index(device_index)
        LOG.debug('   {}'.format(dev['name']))
        if dev['maxInputChannels'] > 0 and pattern.match(dev['name']):
            LOG.debug('    ^-- matched')
            return device_index
    return None



# backwards compat - moved to own python package
from ovos_config.config import Configuration as _Config
from ovos_config.config import MycroftUserConfig, MycroftDefaultConfig, MycroftSystemConfig, RemoteConf, LocalConf
from ovos_config.locations import OLD_USER_CONFIG, DEFAULT_CONFIG, SYSTEM_CONFIG, REMOTE_CONFIG, USER_CONFIG, WEB_CONFIG_CACHE
from ovos_utils.log import log_deprecation


class Configuration(_Config):
    @classmethod
    def get(cls, *args, **kwargs):
        """
        Backwards-compat `get` method from
        https://github.com/MycroftAI/mycroft-core/blob/dev/mycroft/configuration/config.py
        """
        configs = args[0] if len(args) > 0 else kwargs.get("configs", None)
        if configs or isinstance(configs, list):
            log_deprecation("`Configuration.get` now implements `dict.get`",
                            "0.1.0")
            return Configuration.fake_get(*args, **kwargs)

    @staticmethod
    def fake_get(configs=None, cache=True, remote=True):
        """DEPRECATED - use Configuration class instead"""
        # NOTE: this is only called if using the class directly
        # if using an instance (dict object) self._real_get is called instead
        return Configuration.load_config_stack(configs, cache, remote)


# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
NOTE: this is dead code! do not use!

This file is only present to ensure backwards compatibility
in case someone is importing from here

This is only meant for 3rd party code expecting ovos-core
to be a drop in replacement for mycroft-core

"""

from ovos_config.locale import setup_locale
from ovos_config.config import Configuration
from ovos_gui.service import GUIService
from ovos_utils.log import LOG
from ovos_utils import wait_for_exit_signal
from ovos_utils.process_utils import reset_sigint_handler


def on_ready():
    LOG.info("Enclosure started!")


def on_stopping():
    LOG.info('Enclosure is shutting down...')


def on_error(e='Unknown'):
    LOG.error(f'Enclosure failed: {e}')


def create_enclosure(platform):
    """Create an enclosure based on the provided platform string.

    Args:
        platform (str): platform name string

    Returns:
        Enclosure object
    """
    if platform == "mycroft_mark_1":
        LOG.info("Creating Mark I Enclosure")
        LOG.warning("'mycroft_mark_1' enclosure has been deprecated!\n"
                    "'mark_1' support is being migrated into PHAL\n"
                    "see https://github.com/OpenVoiceOS/ovos_phal_mk1")
        from mycroft.deprecated.enclosure.mark1 import EnclosureMark1
        enclosure = EnclosureMark1()
    elif platform == "mycroft_mark_2":
        LOG.info("Creating Mark II Enclosure")
        LOG.warning("'mycroft_mark_2' enclosure has been deprecated!\n"
                    "It was never implemented outside the mk2 feature branch\n"
                    "mark_2 support is being migrated into PHAL\n"
                    "see https://github.com/OpenVoiceOS/ovos_phal_mk2")
        from mycroft.deprecated.enclosure.mark2 import EnclosureMark2
        enclosure = EnclosureMark2()
    else:
        LOG.info("Creating generic enclosure, platform='{}'".format(platform))
        from mycroft.deprecated.enclosure.generic import EnclosureGeneric
        enclosure = EnclosureGeneric()

    return enclosure


def main(ready_hook=on_ready, error_hook=on_error, stopping_hook=on_stopping):
    """Launch one of the available enclosure implementations.

    This depends on the configured platform and can currently either be
    mycroft_mark_1 or mycroft_mark_2, if unconfigured a generic enclosure will be started.

    NOTE: in ovos-core the GUI protocol is handled in it's own service and not part of the enclosure like in mycroft-core!
          You need to also run mycroft.gui process separately, it has been extracted into it's own module
    """
    # Read the system configuration
    config = Configuration()

    LOG.warning("mycroft.client.enclosure is DEPRECATED in ovos-core!")
    LOG.warning("see https://github.com/OpenVoiceOS/ovos_PHAL")

    if not config.get("backwards_compat", True):
        raise DeprecationWarning("Please run PHAL instead of enclosure")

    reset_sigint_handler()
    setup_locale()

    platform = config.get("enclosure", {}).get("platform")

    if platform == "PHAL":
        LOG.debug("Launching PHAL")
        # config read from mycroft.conf
        # "PHAL": {
        #     "ovos-PHAL-plugin-display-manager-ipc": {"enabled": true},
        #     "ovos-PHAL-plugin-mk1": {"enabled": True}
        # }
        try:
            from ovos_PHAL import PHAL
            phal = PHAL()
            phal.start()
            wait_for_exit_signal()
        except Exception as e:
            LOG.exception("PHAL failed to launch!")
            error_hook(e)
    else:
        enclosure = create_enclosure(platform)
        if enclosure:
            LOG.debug("Enclosure created")
            try:
                enclosure.run()
                ready_hook()
            except Exception as e:
                error_hook(e)
        else:
            LOG.info("No enclosure available for this hardware, running headless")

        LOG.warning("Backwards compatibility is enabled, attempting to launch gui service...")
        LOG.warning("Please run PHAL + gui service as separate processes instead!")
        try:
            service = GUIService()
            service.run()
        except Exception as e:
            LOG.error(f"GUI : {e}")
            service = None

        ready_hook()
        wait_for_exit_signal()

        if enclosure:
            enclosure.stop()
        if service:
            service.stop()
        stopping_hook()


if __name__ == "__main__":
    main()



# backwards compat imports
from ovos_audio.tts import TTSFactory
from ovos_plugin_manager.templates.tts import TTS
from ovos_audio.audio import AudioService
from ovos_audio.service import PlaybackService, on_ready, on_alive, on_started, on_error, on_stopping